// (C) Copyright David James Pearce 2003. Permission to copy, use,
// modify, sell and distribute this software is granted provided this
// copyright notice appears in all copies. This software is provided
// "as is" without express or implied warranty, and with no claim as
// to its suitability for any purpose.
//
// Email: d.pearce@doc.ic.ac.uk
//
// Notes
// =====
// This is another solution to the online topological 
// order problem.  It was first described in the 
// following:
//
// B. Alpern, R. Hoover, B. K. Rosen, P. F. Sweeney 
// and F. K. Zadeck, "Incremental Evaluation of
// Computational Circuits", In Proc. of the First 
// Annual ACM-SIAM Symposium on Discrete Algorithms, 
// pages 32-42, 1990

#ifndef AHRSZ_ONLINE_TOPOLOGICAL_ORDER_HPP
#define AHRSZ_ONLINE_TOPOLOGICAL_ORDER_HPP

#include <queue>
#include <vector>
#include <stdexcept>
#include <boost/property_map.hpp>
#include <boost/graph/topological_sort.hpp>

#include "ordered_slist.hpp"
#include "oto_tags.hpp"

#undef AHRSZ_DEBUG

template<class T, class N2I, class I2NMAP>
class ahrsz_online_topological_order;

// --------------------------------
// priority compare 
//
// This gives a comparison based
// on the priority of a node
// --------------------------------

template<class T, class N2I, class Comp>
class ahrsz_priority_comp {
private:
  typedef boost::property_map<T, N2I>::type priority_map;
private:
  Comp _comp;
  priority_map _pmap; // efficiency  
public:
  ahrsz_priority_comp(priority_map pmap)
    : _pmap(pmap) {
  }

  bool operator()(typename T::vertex_descriptor l, 
		  typename T::vertex_descriptor r) const {
    return _comp(_pmap[l],_pmap[r]);
  }
};

// -------------------------
// temporary data associated
// with each visited node
// -------------------------

class ahrsz_node_data {
public:
  bool Marked;
  vector<bool> VisitedForw;
  vector<bool> VisitedBack;
  unsigned int NumEdges;
public:
  ahrsz_node_data(unsigned int n, unsigned int e) 
    : Marked(false),
      VisitedForw(n,false),
      VisitedBack(n,false),
      NumEdges(e) {
  }
};

class ahrsz_priority_value {
private:
  ordered_slist<void>::iterator _iter;
public:
  ahrsz_priority_value(void) 
    : _iter() {
  }

  ahrsz_priority_value(ordered_slist<void>::iterator const &iter) 
  : _iter(iter) {    
  }

  bool operator<(ahrsz_priority_value const &src) const {
    return _iter.base()->first < src._iter.base()->first;
  }

  bool operator>(ahrsz_priority_value const &src) const {
    return _iter.base()->first > src._iter.base()->first;
  }

  bool operator!=(ahrsz_priority_value const &src) const {
    return _iter != src._iter;
  }

  bool operator==(ahrsz_priority_value const &src) const {
    return _iter == src._iter;
  }

  ordered_slist<void>::iterator &base(void) {
    return _iter;
  }

  ahrsz_priority_value const &operator++(int) {
    ++_iter;
    return *this;
  }

  ahrsz_priority_value operator++(void) {
    ahrsz_priority_value r(*this);
    ++_iter;
    return r;
  }
};

// -----------------------
// the discovery algorithm
// -----------------------

template<class T, class N2I, class I2NMAP> 
class ahrsz_discovery {
private:
  // some useful internal typedefs
  typedef boost::property_map<T, N2I>::type priority_map;
  typedef typename T::vertex_descriptor vd_t;
  typedef priority_queue<vd_t,vector<vd_t>,ahrsz_priority_comp<T,N2I,less<ahrsz_priority_value> > > max_priority_queue;
  typedef priority_queue<vd_t,vector<vd_t>,ahrsz_priority_comp<T,N2I,greater<ahrsz_priority_value> > > min_priority_queue;
  typedef typename T::out_edge_iterator out_iterator;
  typedef typename T::in_edge_iterator in_iterator;
private:
  vector<vd_t> &_K; // minimal extended cover
  vector<std::pair<vd_t,vd_t> > &_A;
  set<unsigned int> _Worklist;  

  typedef map<vd_t, ahrsz_node_data> NodeDataMap_t;
  NodeDataMap_t _node_data_map;

  vector<min_priority_queue> _ForwFron;  // the forward frontier
  vector<max_priority_queue> _BackFron;  // the backward frontier

  ahrsz_online_topological_order<T,N2I,I2NMAP> &_g; // the graph!
  boost::property_map<T, N2I>::type _pmap; // efficiency
public:
  ahrsz_discovery(vector<vd_t> &K,
		  vector<std::pair<vd_t,vd_t> > &A,
		  ahrsz_online_topological_order<T,N2I,I2NMAP> &g) 
    : _K(K), _A(A), _ForwFron(_A.size(), min_priority_queue(get(N2I(),g))), 
		    _BackFron(A.size(), max_priority_queue(get(N2I(),g))),
		    _g(g) {
    _pmap = get(N2I(),g);
  }
  
  void operator()(void) {
    for(unsigned int e=0;e!=_A.size();++e) {
      vd_t t(_A[e].first);
      vd_t h(_A[e].second);

      visit(t).VisitedBack[e] = true;
      _BackFron[e].push(t);
      visit(h).VisitedForw[e] = true;
      _ForwFron[e].push(h);

      _Worklist.insert(e);
    }

    while(!_Worklist.empty()) {
      unsigned int e(*_Worklist.begin());
      _Worklist.erase(e);

      ProcessEdge(e);
    }
  }
  
private:
  void ProcessEdge(unsigned int e) {
#ifdef AHRSZ_DEBUG
    cout << "ProcessEdge " << _A[e].first << "->" << _A[e].second << endl;
#endif
    vd_t f,b;

    if(GetTop(_ForwFron[e],f) && GetTop(_BackFron[e],b)) {      
      if(!(_pmap[b] < _pmap[f])) {
	ahrsz_node_data &fd(visit(f));
	ahrsz_node_data &bd(visit(b));
	
	// these nodes are out of order
	unsigned int u = min(bd.NumEdges,fd.NumEdges);
	bd.NumEdges = bd.NumEdges - u;
	fd.NumEdges = fd.NumEdges - u;

	if(bd.NumEdges == 0) { Extend(b,bd); }
	if(fd.NumEdges == 0) { Extend(f,fd); }
      }
    }
  }

  template<class PriorityQueue>
  bool GetTop(PriorityQueue &q, vd_t &r) {
    while(!q.empty()) {
      ahrsz_node_data &dat(visit(q.top()));
      if(!dat.Marked) {
	r = q.top();
	return true;
      }
      q.pop();
    }	
    return false;
  }
  
  void Extend(vd_t x, ahrsz_node_data &xd) {
#ifdef AHRSZ_DEBUG
    cout << "Extend " << x << endl;
#endif
    if(!xd.Marked) {
      _K.push_back(x);
    }
    xd.Marked=true;

    for(unsigned int e=0;e!=_A.size();++e) {
      if(xd.VisitedForw[e]) {
	_Worklist.insert(e);
	GrowForw(e,x);
      }
      if(xd.VisitedBack[e]) {
	_Worklist.insert(e);
	GrowBack(e,x);
      }
    }
  }

  void GrowForw(unsigned int e, vd_t x) {
#ifdef AHRSZ_DEBUG
    cout << "GrowForw " << x << endl;
#endif
    out_iterator i,iend;
    for(tie(i,iend) = out_edges(x,_g);i!=iend;++i) {
      ahrsz_node_data &tmp = visit(target(*i,_g));
      if(!tmp.VisitedForw[e]) {
	tmp.VisitedForw[e]=true;
	if(tmp.Marked) {	  
	  GrowForw(e,target(*i,_g));
	  // this recursive call doesn't invalidate
	  // tmp, because of std::map's iterator
	  // properties
	  // tmp = visit(*i);
	} else {
#ifdef AHRSZ_DEBUG
	  cout << "ForwFron.push(" << target(*i,_g) << ")" << endl;
#endif
	  _ForwFron[e].push(target(*i,_g));
	}
      }
    }
  }

  void GrowBack(unsigned int e, vd_t x) {
#ifdef AHRSZ_DEBUG
    cout << "GrowBack " << x << endl;
#endif
    in_iterator i,iend;
    for(tie(i,iend) = in_edges(x,_g);i!=iend;++i) {
      ahrsz_node_data &tmp = visit(source(*i,_g));
      if(!tmp.VisitedBack[e]) {
	tmp.VisitedBack[e]=true;
	if(tmp.Marked) {
	  GrowBack(e,source(*i,_g));
	  // this recursive call doesn't invalidate
	  // tmp, because of std::map's iterator
	  // properties
	  // tmp = visit(*i);
	} else {
#ifdef AHRSZ_DEBUG
	  cout << "BackFron.push(" << source(*i,_g) << ")" << endl;
#endif
	  _BackFron[e].push(source(*i,_g));
	}
      }
    }
  }
  
  ahrsz_node_data &visit(vd_t x) {
    NodeDataMap_t::iterator i(_node_data_map.lower_bound(x));

    if(i == _node_data_map.end() || x < i->first) {
      // node not already visited
      i=_node_data_map.insert(i, NodeDataMap_t::value_type(x,ahrsz_node_data
							   (_A.size(), in_degree(x,_g) 
							    + out_degree(x,_g))));      
    }
    
    return i->second;
  }
};

// --------------------------
// the reassignment algorithm
// --------------------------

template<class T, class N2I, class I2NMAP>
class ahrsz_reassign {
private:
  class node_data {
  public:
    bool visited;
    ahrsz_priority_value ceiling;
    //    typename T::vertex_size_type degree;
  public:
    node_data(void) {
    }
  };
private:
  typedef typename T::vertex_descriptor vd_t;
private:
  vector<vd_t> &_K;
  map<vd_t,unsigned int> _rK;
  vector<node_data> _data;
  boost::property_map<T, N2I>::type _pmap; // efficiency
  ahrsz_online_topological_order<T,N2I,I2NMAP> &_g;
  vector<vd_t> _reverse_topological_order;
public:
  ahrsz_reassign(vector<vd_t> &K, 
		 ahrsz_online_topological_order<T,N2I,I2NMAP> &g) 
    : _K(K), _data(_K.size()), _g(g) {
    for(unsigned int i=0;i!=_K.size();++i) {
      _rK[_K[i]] = i;
      _data[i].ceiling = ahrsz_priority_value(g._pspace.end());
      //      _data[i].degree = in_degree(_K[i],g);
      _data[i].visited = false;
    }

    _pmap = get(N2I(),g);
  }

  void operator()(void) {

    // first pass - compute ceiling information
    unsigned int index(0);
    for(vector<vd_t>::iterator i(_K.begin());
	i!=_K.end();++i, index++) {
      if(!_data[index].visited) {
	compute_ceilings(*i);
      }
    }

    // second pass - perform the reassignment
    for(vector<vd_t>::reverse_iterator i(_reverse_topological_order.rbegin());
	i!=_reverse_topological_order.rend();++i) {
      bool minus_inf(true);
      ahrsz_priority_value floor;
      
      typename T::in_edge_iterator j,jend;
      for(tie(j,jend) = in_edges(*i,_g);j!=jend;++j) {
	if(minus_inf || _pmap[source(*j,_g)] > floor) {
	  floor = _pmap[source(*j,_g)];
	  minus_inf = false;
	}
      }

      // right, so we're good to assign a priority
      // which lies between [but does not equal] 
      // floor and ceiling 
      node_data &data(_data[_rK[*i]]);
      bool plus_inf(data.ceiling == _g._pspace.end());

#ifdef AHRSZ_DEBUG
      cout << "REASSIGN " << *i << ", Ceiling = ";
      if(plus_inf) { cout << "+inf"; }
      else { cout << data.ceiling.base().base()->first; }
      cout << ", Floor = ";
      if(minus_inf) { cout << "-inf"; }
      else { cout << floor.base().base()->first; }
#endif

      if((minus_inf || floor < _pmap[*i]) && (plus_inf || _pmap[*i] < data.ceiling)) {
	// don't actually need to do anything here
	// as the priority is already valid...
#ifdef AHRSZ_DEBUG
	cout << ", prioryt left at " << _pmap[*i].base().base()->first << endl;
#endif
      } else {
	ahrsz_priority_value candidate;

	if(minus_inf) { 
	  candidate = _g._pspace.begin();
	} else {
	  candidate = floor;
	  ++candidate;
	}

	if(candidate == _g._pspace.end() || !(plus_inf || candidate < data.ceiling)) {
	  // there is no priority after floor OR candidate
	  // is greater than the ceiling
	  // eitherway, we must add a new priority ...365
	  if(minus_inf) {
	    // this is dodgy. surely we could
	    // use _pspace.begin() already ?
	    _g._pspace.push_front();
	    _pmap[*i] = _g._pspace.begin();
	  } else {
	    _pmap[*i] = _g._pspace.insert_after(floor.base());
	  }
	} else {
	  // candidate is good
	  _pmap[*i] = candidate;
	} 
#ifdef AHRSZ_DEBUG
	cout << ", new priority " << _pmap[*i].base().base()->first << endl;
#endif
      }
    }
  }

private:
  void compute_ceilings(vd_t n) {
    node_data &data(_data[_rK[n]]);
    data.visited=true;

    typename T::out_edge_iterator i,iend;
    for(tie(i,iend) = out_edges(n,_g);i!=iend;++i) {
      // could eliminate this _rK call by passing
      // it as a parameter to compute_ceilings      
      if(_rK.find(target(*i,_g)) != _rK.end()) {
	if(!_data[_rK[target(*i,_g)]].visited) {
	  // node not yet visited and in K
	  compute_ceilings(target(*i,_g));
	}
	node_data &sdata(_data[_rK[target(*i,_g)]]);
	if(sdata.ceiling != _g._pspace.end() && 
	   (data.ceiling == _g._pspace.end() ||
	    sdata.ceiling < data.ceiling)) {
	  data.ceiling = sdata.ceiling;
	}
      } else if(data.ceiling == _g._pspace.end() 
		|| _pmap[target(*i,_g)] < data.ceiling) {
	data.ceiling = _pmap[target(*i,_g)];
      }       
    }
    _reverse_topological_order.push_back(n);
  }
  };

// --------------
// the main class
// --------------

template<class T, class N2I, class I2NMAP>
std::pair<typename T::edge_descriptor, bool> 
add_edge(typename ahrsz_online_topological_order<T,N2I,I2NMAP>::vertex_descriptor t, 
	 typename ahrsz_online_topological_order<T,N2I,I2NMAP>::vertex_descriptor h, 
	 ahrsz_online_topological_order<T,N2I,I2NMAP> &g) {

#ifdef AHRSZ_DEBUG
  cout << "ADD_EDGE " << t << "->" << h << endl;
#endif

  std::pair<typename T::edge_descriptor, bool> r;
  r = add_edge(static_cast<typename T::vertex_descriptor>(t),
	       static_cast<typename T::vertex_descriptor>(h),
	       static_cast<T&>(g));

  if(r.second) {
    // the edge is indeed new
    vector<typename T::vertex_descriptor> K;
    vector<std::pair<unsigned int, unsigned int> > A;
    A.push_back(std::make_pair(t,h));
    
    // do the discovery
    ahrsz_discovery<T,N2I,I2NMAP> d(K, A, g);  
    d();

#ifdef AHRSZ_DEBUG
    cout << "COVER: ";
    for(unsigned int i=0;i!=K.size();++i) {
    cout << K[i] << " ";
    }
    cout << endl;
#endif

    // now to the reassignment
    ahrsz_reassign<T,N2I,I2NMAP> re(K,g);
    re();
    // all done.
  }
}

template<class T, class N2I = n2i_t, class I2NMAP = vector<typename T::vertex_descriptor> >
class ahrsz_online_topological_order : public T {
public:
  typedef ahrsz_online_topological_order<T> self;
  friend std::pair<typename T::edge_descriptor, bool> add_edge<T>(typename self::vertex_descriptor, 
								  typename self::vertex_descriptor, 
								  self &);
  friend class ahrsz_reassign<T,N2I,I2NMAP>;
private:
  I2NMAP _i2n;
  ordered_slist<void> _pspace; // the priority space.
public:
  ahrsz_online_topological_order(T const &g) 
    : T(g), _i2n(num_vertices(g)), _pspace(1) {    
    boost::property_map<T, N2I>::type _pmap = get(N2I(),*this);
    
    typename T::vertex_iterator i,iend;
    for(tie(i,iend) = vertices(*this);i!=iend;++i) {
      _pmap[*i] = ahrsz_priority_value(_pspace.begin());
    }
  }

  ahrsz_online_topological_order(typename T::vertices_size_type n) 
    : T(n), _i2n(n), _pspace(1) {
    boost::property_map<T, N2I>::type _pmap = get(N2I(),*this);
    
    typename T::vertex_iterator i,iend;
    for(tie(i,iend) = vertices(*this);i!=iend;++i) {
      _pmap[*i] = ahrsz_priority_value(_pspace.begin());
    }
  }

private:
  ahrsz_online_topological_order(ahrsz_online_topological_order const &src) {
    // ------------------------------------------------------------------
    // this is private for the moment because it needs special attention
    // and I haven't got around to doing it yet. In particular, the 
    // _pspace iterators must be retargeted to point into the new _pspace
    // ------------------------------------------------------------------
  }
};

#endif



