// (C) Copyright David James Pearce 2003. Permission to copy, use,
// modify, sell and distribute this software is granted provided this
// copyright notice appears in all copies. This software is provided
// "as is" without express or implied warranty, and with no claim as
// to its suitability for any purpose.
//
// Email: d.pearce@doc.ic.ac.uk
//
// Notes
// =====
// This is another solution to the online topological 
// order problem.  It was first described in the 
// following:
//
// B. Alpern, R. Hoover, B. K. Rosen, P. F. Sweeney 
// and F. K. Zadeck, "Incremental Evaluation of
// Computational Circuits", In Proc. of the First 
// Annual ACM-SIAM Symposium on Discrete Algorithms, 
// pages 32-42, 1990

#ifndef AHRSZ_ONLINE_TOPOLOGICAL_ORDER2_HPP
#define AHRSZ_ONLINE_TOPOLOGICAL_ORDER2_HPP

#include <queue>
#include <vector>
#include <stdexcept>
#include <functional>
#include <boost/property_map.hpp>
#include <boost/graph/topological_sort.hpp>

#include "ordered_slist.hpp"
#include "oto_tags.hpp"

#if __GNUC__ >= 3
#include <ext/functional>
// #include <limits>
// #define std::numeric_limits<unsigned int>::max() OL_UINT_MAX
using __gnu_cxx::identity;
using __gnu_cxx::select1st;
#else
using std::identity;
using std::select1st;
#endif

#undef AHRSZ_DEBUG

namespace experimental {

  template<class T, class S = identity<T> >
  class sless {
  private:
    S _s;
  public:
    inline bool operator()(T const &a, T const &b) const {
      return _s(a) < _s(b);
    }
  };

template<class T, class N2I, class I2NMAP>
class ahrsz_online_topological_order;

// --------------------------------
// priority compare 
//
// This gives a comparison based
// on the priority of a node
// --------------------------------

template<class T, class N2I, class Comp>
class ahrsz_priority_comp {
private:
  typedef typename boost::property_map<T, N2I>::type priority_map;
private:
  Comp _comp;
  priority_map _pmap; // efficiency  
public:
  ahrsz_priority_comp(priority_map pmap)
    : _pmap(pmap) {
  }

  bool operator()(typename T::vertex_descriptor l, 
		  typename T::vertex_descriptor r) const {
    return _comp(_pmap[l],_pmap[r]);
  }
};

// -------------------------
// temporary data associated
// with each visited node
// -------------------------

class ahrsz_node_data {
public:
  bool Marked;
  std::vector<bool> VisitedForw;
  std::vector<bool> VisitedBack;
  unsigned int NumEdges;
public:
  ahrsz_node_data(unsigned int n, unsigned int e) 
    : Marked(false),
    VisitedForw(n,false),
    VisitedBack(n,false),
    NumEdges(e) {
  }
};

class ahrsz_priority_value {
private:
  ordered_slist<void>::iterator _iter;
public:
  ahrsz_priority_value(void) 
    : _iter() {
  }

  ahrsz_priority_value(ordered_slist<void>::iterator const &iter) 
  : _iter(iter) {    
  }

  bool operator<(ahrsz_priority_value const &src) const {
    return _iter.base()->first < src._iter.base()->first;
  }

  bool operator>(ahrsz_priority_value const &src) const {
    return _iter.base()->first > src._iter.base()->first;
  }

  bool operator!=(ahrsz_priority_value const &src) const {
    return _iter != src._iter;
  }

  bool operator==(ahrsz_priority_value const &src) const {
    return _iter == src._iter;
  }

  ordered_slist<void>::iterator &base(void) {
    return _iter;
  }

  ahrsz_priority_value const &operator++(int) {
    ++_iter;
    return *this;
  }

  ahrsz_priority_value operator++(void) {
    ahrsz_priority_value r(*this);
    ++_iter;
    return r;
  }
};

  typedef enum {minus_infinity,plus_infinity} ahrsz_ext_priority_value_inf_t;
  
  class ahrsz_ext_priority_value {
  private:
    ordered_slist<void>::iterator _iter;
    bool _minus_inf;
    bool _plus_inf;
public:
  ahrsz_ext_priority_value(ahrsz_ext_priority_value_inf_t f = minus_infinity) 
    : _iter(), _minus_inf(f == minus_infinity), _plus_inf(f == plus_infinity) {    
  }

  ahrsz_ext_priority_value(ahrsz_priority_value &e) 
    : _iter(e.base()), _minus_inf(false), _plus_inf(false) {    
  }

  ahrsz_ext_priority_value(ordered_slist<void>::iterator const &iter) 
    : _iter(iter), _minus_inf(false), _plus_inf(false) {    
  }

  bool minus_inf(void) const { return _minus_inf; }
  bool plus_inf(void) const { return _plus_inf; }

  bool operator<(ahrsz_ext_priority_value const &src) const {
    if(_minus_inf || src._minus_inf) {
      return _minus_inf && !src._minus_inf;
    } else if(_plus_inf || src._plus_inf) {
      return src._plus_inf && !_plus_inf;
    } 
    return _iter.base()->first < src._iter.base()->first;
  }

  bool operator>=(ahrsz_ext_priority_value const &src) const {
    return src < *this;
  }

  bool operator>(ahrsz_ext_priority_value const &src) const {
    return src < (*this);
  }

  bool operator==(ahrsz_ext_priority_value const &src) const {
    if(_minus_inf || src._minus_inf) {
      return _minus_inf == src._minus_inf;
    } else if(_plus_inf || src._plus_inf) {
      return _plus_inf == src._plus_inf;
    }
    return _iter == src._iter;
  }

  bool operator!=(ahrsz_ext_priority_value const &src) const {
    return !(*this == src);
  }

  ordered_slist<void>::iterator &base(void) {
    assert(!_minus_inf && !_plus_inf);
    return _iter;
  }

  ahrsz_ext_priority_value const &operator++(int) {
    ++_iter;
    return *this;
  }

  ahrsz_ext_priority_value operator++(void) {
    ahrsz_ext_priority_value r(*this);
    ++_iter;
    return r;
  }
};

// -----------------------
// the discovery algorithm
// -----------------------

template<class T, class N2I, class I2NMAP> 
class ahrsz_discovery {
private:
  // some useful internal typedefs
  typedef typename boost::property_map<T, N2I>::type priority_map;
  typedef typename T::vertex_descriptor vd_t;
  typedef std::priority_queue<vd_t,std::vector<vd_t>,
			      ahrsz_priority_comp<T,N2I,std::less<ahrsz_priority_value> > > max_priority_queue;
  typedef std::priority_queue<vd_t,std::vector<vd_t>,
			      ahrsz_priority_comp<T,N2I,std::greater<ahrsz_priority_value> > > min_priority_queue;
  typedef typename T::out_edge_iterator out_iterator;
  typedef typename T::in_edge_iterator in_iterator;
private:
  std::vector<vd_t> &_K; // minimal extended cover
  std::vector<std::pair<vd_t,vd_t> > &_A;
  std::set<unsigned int> _Worklist;  

  typedef std::map<vd_t, ahrsz_node_data> NodeDataMap_t;
  NodeDataMap_t _node_data_map;

  std::vector<min_priority_queue> _ForwFron;  // the forward frontier
  std::vector<max_priority_queue> _BackFron;  // the backward frontier

  ahrsz_online_topological_order<T,N2I,I2NMAP> &_g; // the graph!
  typename boost::property_map<T, N2I>::type _pmap; // efficiency
public:
  ahrsz_discovery(std::vector<vd_t> &K,
		  std::vector<std::pair<vd_t,vd_t> > &A,
		  ahrsz_online_topological_order<T,N2I,I2NMAP> &g) 
    : _K(K), _A(A), _ForwFron(_A.size(), min_priority_queue(get(N2I(),g))), 
		    _BackFron(A.size(), max_priority_queue(get(N2I(),g))),
		    _g(g) {
    _pmap = get(N2I(),g);
  }
  
  void operator()(void) {
    for(unsigned int e=0;e!=_A.size();++e) {
      vd_t t(_A[e].first);
      vd_t h(_A[e].second);

      visit(t).VisitedBack[e] = true;
      _BackFron[e].push(t);
      visit(h).VisitedForw[e] = true;
      _ForwFron[e].push(h);

      _Worklist.insert(e);
    }

    while(!_Worklist.empty()) {
      unsigned int e(*_Worklist.begin());
      _Worklist.erase(e);

      ProcessEdge(e);
    }
  }
  
private:
  void ProcessEdge(unsigned int e) {
#ifdef AHRSZ_DEBUG
    cout << "ProcessEdge " << _A[e].first << "->" << _A[e].second << endl;
#endif
    vd_t f,b;

    if(GetTop(_ForwFron[e],f) && GetTop(_BackFron[e],b)) {      
      if(!(_pmap[b] < _pmap[f])) {
	ahrsz_node_data &fd(visit(f));
	ahrsz_node_data &bd(visit(b));
	
	// these nodes are out of order
	unsigned int u = min(bd.NumEdges,fd.NumEdges);
	bd.NumEdges = bd.NumEdges - u;
	fd.NumEdges = fd.NumEdges - u;

	if(bd.NumEdges == 0) { Extend(b,bd); }
	if(fd.NumEdges == 0) { Extend(f,fd); }
      }
    }
  }

  template<class PriorityQueue>
  bool GetTop(PriorityQueue &q, vd_t &r) {
    while(!q.empty()) {
      ahrsz_node_data &dat(visit(q.top()));
      if(!dat.Marked) {
	r = q.top();
	return true;
      }
      q.pop();
    }	
    return false;
  }
  
  void Extend(vd_t x, ahrsz_node_data &xd) {
#ifdef AHRSZ_DEBUG
    cout << "Extend " << x << endl;
#endif
    if(!xd.Marked) {
      _K.push_back(x);
    }
    xd.Marked=true;

    for(unsigned int e=0;e!=_A.size();++e) {
      if(xd.VisitedForw[e]) {
	_Worklist.insert(e);
	GrowForw(e,x);
      }
      if(xd.VisitedBack[e]) {
	_Worklist.insert(e);
	GrowBack(e,x);
      }
    }
  }

  void GrowForw(unsigned int e, vd_t x) {
#ifdef AHRSZ_DEBUG
    cout << "GrowForw " << x << endl;
#endif
    out_iterator i,iend;
    for(tie(i,iend) = out_edges(x,_g);i!=iend;++i) {
      ahrsz_node_data &tmp = visit(target(*i,_g));
      if(!tmp.VisitedForw[e]) {
	tmp.VisitedForw[e]=true;
	if(tmp.Marked) {	  
	  GrowForw(e,target(*i,_g));
	  // this recursive call doesn't invalidate
	  // tmp, because of std::map's iterator
	  // properties
	  // tmp = visit(*i);
	} else {
#ifdef AHRSZ_DEBUG
	  cout << "ForwFron.push(" << target(*i,_g) << ")" << endl;
#endif
	  _ForwFron[e].push(target(*i,_g));
	}
      }
    }
  }

  void GrowBack(unsigned int e, vd_t x) {
#ifdef AHRSZ_DEBUG
    cout << "GrowBack " << x << endl;
#endif
    in_iterator i,iend;
    for(tie(i,iend) = in_edges(x,_g);i!=iend;++i) {
      ahrsz_node_data &tmp = visit(source(*i,_g));
      if(!tmp.VisitedBack[e]) {
	tmp.VisitedBack[e]=true;
	if(tmp.Marked) {
	  GrowBack(e,source(*i,_g));
	  // this recursive call doesn't invalidate
	  // tmp, because of std::map's iterator
	  // properties
	  // tmp = visit(*i);
	} else {
#ifdef AHRSZ_DEBUG
	  cout << "BackFron.push(" << source(*i,_g) << ")" << endl;
#endif
	  _BackFron[e].push(source(*i,_g));
	}
      }
    }
  }
  
  ahrsz_node_data &visit(vd_t x) {
    typename NodeDataMap_t::iterator i(_node_data_map.lower_bound(x));

    if(i == _node_data_map.end() || x < i->first) {
      // node not already visited
      i=_node_data_map.insert(i, typename NodeDataMap_t::value_type(x,ahrsz_node_data
							   (_A.size(), in_degree(x,_g) 
							    + out_degree(x,_g))));      
    }
    
    return i->second;
  }
};

// --------------------------
// the reassignment algorithm
// --------------------------

template<class T, class N2I, class I2NMAP>
class ahrsz_reassign {
private:
  class node_data {
  public:
    bool visited;
    ahrsz_ext_priority_value ceiling;
    typename T::degree_size_type degree;
  public:
    node_data(void) : ceiling(plus_infinity) {
    }
  };
private:
  typedef typename T::vertex_descriptor vd_t;
private:
  typedef std::map<vd_t,unsigned int> rK_t;
  std::vector<vd_t> &_K;
  rK_t _rK;
  std::vector<node_data> _data;
  typename boost::property_map<T, N2I>::type _pmap; // efficiency
  ahrsz_online_topological_order<T,N2I,I2NMAP> &_g;
  //  std::vector<vd_t> _reverse_topological_order;
public:
  ahrsz_reassign(std::vector<vd_t> &K, 
		 ahrsz_online_topological_order<T,N2I,I2NMAP> &g) 
    : _K(K), _data(_K.size()), _g(g) {
    for(unsigned int i=0;i!=_K.size();++i) {
      _rK[_K[i]] = i;
    }
    // necessary because we must know what is in K
    for(unsigned int i=0;i!=_K.size();++i) {
      _data[i].degree = compute_in_degree(_K[i]);
      _data[i].visited = false;
    }

    _pmap = get(N2I(),g);
  }

  void operator()(void) {

    // first pass - compute ceiling information
    unsigned int index(0);
    for(typename std::vector<vd_t>::iterator i(_K.begin());
	i!=_K.end();++i, index++) {
      if(!_data[index].visited) {
	compute_ceilings(*i,_data[index]);
      }
    }

    // second pass - perform the reassignment
    typedef std::pair<ahrsz_ext_priority_value,vd_t> p_t;
    typedef std::multiset<p_t,sless<p_t,select1st<p_t> > > pqueue_t;
    pqueue_t pq;
    
    // initialise the priority queue with all vertices with 
    // no predecessor in K.

    for(typename std::vector<vd_t>::iterator i(_K.begin());i!=_K.end();++i) {
      bool predInK(false);
      typename T::in_edge_iterator j,jend;      
      for(tie(j,jend)=in_edges(*i,_g);j!=jend;++j) {
	if(_rK.find(source(*j,_g)) != _rK.end()) {
	  predInK=true;
	  break;
	}
      }
      if(!predInK) {
	pq.insert(std::make_pair(compute_floor(*i),*i));
      }
    }
    
    // now, do the loop i guess.
    while(!pq.empty()) {
      typename pqueue_t::value_type n(*pq.begin());
      ahrsz_ext_priority_value ceiling(_data[_rK[n.second]].ceiling);
      typename pqueue_t::iterator iend(pq.begin());
      ++iend;
      while(iend!=pq.end() && iend->first == n.first) {
	ceiling=min(_data[_rK[iend->second]].ceiling,ceiling);
	++iend;
      }      
      
      // at this point we have all vertices with same
      // floor and ceiling is the min ceiling.
      // we must determine the correct value and
      // assign it to all in the range and then pop them
      
      ceiling = compute_priority(n.first,ceiling);
      
      for(typename pqueue_t::iterator i(pq.begin());i!=iend;++i) {
	_pmap[i->second] = ahrsz_priority_value(ceiling.base());
	// decrement degree for each successor
	typename T::out_edge_iterator j,jend;
	for(tie(j,jend) = out_edges(i->second,_g);j!=jend;++j) {
	  typename rK_t::iterator k(_rK.find(target(*j,_g)));
	  if(k != _rK.end() && --_data[k->second].degree == 0) {
	    // I think we can safely assume that this new element
	    // will not be inserted between [pq.begin(),iend);
	    pq.insert(std::make_pair(compute_floor(target(*j,_g)),target(*j,_g)));	    
	  }
	}	
      }
      
      // "pop" nodes off queue ...
      pq.erase(pq.begin(),iend);
    }
  }

  typename T::degree_size_type compute_in_degree(vd_t n) {
    typename T::degree_size_type r(0);
    
    typename T::in_edge_iterator i,iend;
    for(tie(i,iend) = in_edges(n,_g);i!=iend;++i) {
      if(_rK.find(source(*i,_g)) != _rK.end()) {
	++r;
      }
    }
    return r;
  }
  
  ahrsz_ext_priority_value compute_priority(ahrsz_ext_priority_value &floor, 
					    ahrsz_ext_priority_value &ceiling) {
    ahrsz_ext_priority_value candidate;
    
    if(floor.minus_inf()) {
      candidate = _g._pspace.begin();
      if(candidate >= ceiling) {
	// we need a new priority
	_g._pspace.push_front();
	candidate=_g._pspace.begin();
      }
    } else {
      assert(!floor.plus_inf());
      candidate=floor;
      ++candidate;
      if(candidate.base() == _g._pspace.end() || candidate >= ceiling) {
	// we need a new priority
	candidate = _g._pspace.insert_after(floor.base());
      }
    }
    return candidate;
  }
  
  ahrsz_ext_priority_value compute_floor(vd_t v) {
    ahrsz_ext_priority_value floor(minus_infinity);
    
    typename T::in_edge_iterator j,jend;
    for(tie(j,jend) = in_edges(v,_g);j!=jend;++j) {
      floor = std::max(floor,ahrsz_ext_priority_value(_pmap[source(*j,_g)]));
    }
    
    return floor;
  }
  
private:
  void compute_ceilings(vd_t n, node_data &data) {
    data.visited=true;

    typename T::out_edge_iterator i,iend;
    for(tie(i,iend) = out_edges(n,_g);i!=iend;++i) {
      // could eliminate this _rK call by passing
      // it as a parameter to compute_ceilings      
      typename rK_t::iterator j(_rK.find(target(*i,_g)));
      if(j != _rK.end()) {
	// successor is in K
	node_data &sdata(_data[j->second]);
	if(!_data[j->second].visited) {
	  compute_ceilings(target(*i,_g),sdata);
	}
	data.ceiling = std::min(data.ceiling,sdata.ceiling);
      } else {
	// successor is not in K
	data.ceiling = std::min(data.ceiling,ahrsz_ext_priority_value(_pmap[target(*i,_g)]));
      }       
    }
    //    _reverse_topological_order.push_back(n);
  }
  };

// --------------
// the main class
// --------------

template<class T, class N2I, class I2NMAP>
std::pair<typename T::edge_descriptor, bool> 
add_edge(typename ahrsz_online_topological_order<T,N2I,I2NMAP>::vertex_descriptor t, 
	 typename ahrsz_online_topological_order<T,N2I,I2NMAP>::vertex_descriptor h, 
	 ahrsz_online_topological_order<T,N2I,I2NMAP> &g) {

#ifdef AHRSZ_DEBUG
  cout << "ADD_EDGE " << t << "->" << h << endl;
#endif

  std::pair<typename T::edge_descriptor, bool> r;
  r = add_edge(static_cast<typename T::vertex_descriptor>(t),
	       static_cast<typename T::vertex_descriptor>(h),
	       static_cast<T&>(g));

  if(r.second) {
    // the edge is indeed new
    typedef typename boost::property_map<T, N2I>::type N2iMap;
    N2iMap n2i = get(N2I(),g);
    
    if(!(n2i[t] < n2i[h])) {
      std::vector<typename T::vertex_descriptor> K;
      std::vector<std::pair<unsigned int, unsigned int> > A;
      A.push_back(std::make_pair(t,h));
      
      // do the discovery
      ahrsz_discovery<T,N2I,I2NMAP> d(K, A, g);  
      d();
      
#ifdef AHRSZ_DEBUG
      cout << "COVER: ";
      for(unsigned int i=0;i!=K.size();++i) {
	cout << K[i] << " ";
      }
      cout << endl;
#endif
      
      // now to the reassignment
      ahrsz_reassign<T,N2I,I2NMAP> re(K,g);
      re();
      // all done.
    }
  }
}

template<class InputIter, class T, class P, class N2I>
void add_edges(InputIter b, InputIter e, 
	       ahrsz_online_topological_order<T,P,N2I> &g) {
  for(;b!=e;++b) {
    add_edge(b->first,b->second,g);
  }
}

template<class T, class N2I = n2i_t, class I2NMAP = std::vector<typename T::vertex_descriptor> >
class ahrsz_online_topological_order : public T {
public:
  typedef ahrsz_online_topological_order<T> self;
  friend std::pair<typename T::edge_descriptor, bool> add_edge<T>(typename self::vertex_descriptor, 
								  typename self::vertex_descriptor, 
								  self &);
  friend class ahrsz_reassign<T,N2I,I2NMAP>;
private:
  I2NMAP _i2n;
  ordered_slist<void> _pspace; // the priority space.
public:
  ahrsz_online_topological_order(T const &g, unsigned int acc = 1) 
    : T(g), _i2n(num_vertices(g)), _pspace(1) {    
    typename boost::property_map<T, N2I>::type _pmap = get(N2I(),*this);
    
    typename T::vertex_iterator i,iend;
    for(tie(i,iend) = vertices(*this);i!=iend;++i) {
      _pmap[*i] = ahrsz_priority_value(_pspace.begin());
    }
  }

  ahrsz_online_topological_order(typename T::vertices_size_type n, 
				 unsigned int acc = 1) 
    : T(n), _i2n(n), _pspace(1) {
    typename boost::property_map<T, N2I>::type _pmap = get(N2I(),*this);
    
    typename T::vertex_iterator i,iend;
    for(tie(i,iend) = vertices(*this);i!=iend;++i) {
      _pmap[*i] = ahrsz_priority_value(_pspace.begin());
    }
  }

private:
  ahrsz_online_topological_order(ahrsz_online_topological_order const &src) {
    // ------------------------------------------------------------------
    // this is private for the moment because it needs special attention
    // and I haven't got around to doing it yet. In particular, the 
    // _pspace iterators must be retargeted to point into the new _pspace
    // ------------------------------------------------------------------
  }
};
}

#endif



