// (C) Copyright David James Pearce 2003. Permission to copy, use,
// modify, sell and distribute this software is granted provided this
// copyright notice appears in all copies. This software is provided
// "as is" without express or implied warranty, and with no claim as
// to its suitability for any purpose.
//
// Email: d.pearce@doc.ic.ac.uk
//
// Notes
// =====
// This is a wrapper for the incremental connected components
// algorithms which are part of the boost graph library.
//
// I've made this in order to make it easier for someone to
// use generic algorithms with a graph maintaining connected
// components.

#ifndef INC_CONNECTED_COMPONENTS_HPP
#define INC_CONNECTED_COMPONENTS_HPP

#include <boost/pending/disjoint_sets.hpp>
#include <boost/graph/incremental_components.hpp>

template<class T, class V>
class inc_connected_components;

template<class T, class V>
bool same_component(typename inc_connected_components<T,V>::vertex_descriptor t, 
		    typename inc_connected_components<T,V>::vertex_descriptor h, 
		    inc_connected_components<T,V> &g) {
  return same_component(t,h,*g._pds);
}

template<class T, class V>
std::pair<typename T::edge_descriptor, bool> 
add_edge(typename inc_connected_components<T,V>::vertex_descriptor t, 
	 typename inc_connected_components<T,V>::vertex_descriptor h, 
	 inc_connected_components<T,V> &g) {

  std::pair<typename T::edge_descriptor, bool> r;
  r = add_edge(static_cast<typename T::vertex_descriptor>(t),
	       static_cast<typename T::vertex_descriptor>(h),
	       static_cast<T&>(g));
  
  if(r.second) {
    // update the disjoint sets data structure
    g._pds->union_set(t,h);
  }

  // that's it!
  return r;
}

template<class T, class V = vector<typename T::vertex_descriptor> >
class inc_connected_components : public T {
private:
  // friends
  typedef inc_connected_components<T,V> self;
  friend std::pair<typename T::edge_descriptor, bool> add_edge<T>(typename self::vertex_descriptor, 
								  typename self::vertex_descriptor,
								  self &);

  friend bool same_component<T,V>(typename self::vertex_descriptor, 
				  typename self::vertex_descriptor,
				  self &);

  
private:
  // properties
  typedef typename V::iterator Rank;
  typedef typename V::iterator Parent;
  
  V _rank; // could be reference
  V _parent; // could be reference
  boost::disjoint_sets<Rank, Parent> *_pds;
public:
  inc_connected_components(typename T::vertices_size_type n)
    : T(n) { init(); } // init(); }

  inc_connected_components(T const &g) 
    : T(g) { init(); }
  ~inc_connected_components(void) {
    delete _pds;
  }

  inc_connected_components(self const &t) 
    : T(t) {
    // left private coz I don't know the best
    // way to implement this yet. ?
    copy(t);
  }

  self const &operator=(self const &src) {
    if(this != &src) {
      T::operator=(src); // does this work ??
      delete _pds;
      copy(src);
    }
    return *this;
  }
private:  
  void copy(self const &src) {
    _rank = src._rank;
    _parent = src._parent;

    _pds = new boost::disjoint_sets<Rank,Parent>(_rank.begin(),_parent.begin());
  }

  void init(void) {
    _rank.resize(num_vertices(*this));
    _parent.resize(num_vertices(*this));

    _pds = new boost::disjoint_sets<Rank,Parent>(_rank.begin(),_parent.begin());

    boost::initialize_incremental_components(*this, *_pds);
    boost::incremental_components(*this, *_pds);
  }
}; 

#endif
