#ifndef PKH_ONLINE_TOPOLOGICAL_ORDER_HPP
#define PKH_ONLINE_TOPOLOGICAL_ORDER_HPP

#include <stdexcept>
#include <boost/property_map.hpp>
#include <boost/graph/topological_sort.hpp>
#include "oto_tags.hpp"

template<class T, class N2I, class I2NMAP>
class pkh_online_topological_order;

template<class T, class N2I, class I2NMAP> 
bool pkh_oto_dfs(typename T::vertex_descriptor h, 
		 typename T::vertex_descriptor t,
		 std::vector<unsigned int> &reachable,
		 std::vector<unsigned int> &reaching,
		 typename boost::property_map<T, N2I>::type &n2i,		 
		 pkh_online_topological_order<T,N2I,I2NMAP> &g) {
  
  // some useful typedefs
  
  typedef typename T::vertex_descriptor vd_t;
  typedef typename T::out_edge_iterator out_iterator;
  typedef typename T::in_edge_iterator in_iterator;

  // now the actual code.
  // create worklist and guess at
  // likely max size

  unsigned int lb(n2i[h]);
  unsigned int ub(n2i[t]);

  std::vector<vd_t> worklist;    

  // mark node h as visited
  
  g._visited[h]=true;
  worklist.push_back(h);
  reachable.push_back(lb);

  // now perform the forward depth-first-search

  while(worklist.size() > 0) { 
    vd_t n(worklist.back());
    worklist.pop_back();

    out_iterator i,iend;
    for(tie(i,iend) = out_edges(n,g); i!=iend;++i) {
      unsigned int w(target(*i,g));
      unsigned int wn2i(n2i[w]);
      if(wn2i == ub) {
	// this is the special case
	// where a cycle has been detected
	return true;
      } else if(wn2i < ub && !g._visited[w]) {
	g._visited[w]=true;
	worklist.push_back(w);
	reachable.push_back(wn2i);
      }
    }
  }

  // Now, perform the reverse depth-first-search
  worklist.push_back(t);
  g._visited[t] = true;
  reaching.push_back(ub);

  while(worklist.size() > 0) { 
    vd_t n(worklist.back());
    worklist.pop_back();

    in_iterator i,iend;
    for(tie(i,iend) = in_edges(n,g); i!=iend;++i) {
      unsigned int w(source(*i,g));
      unsigned int wn2i(n2i[w]);
       if(wn2i > lb && !g._visited[w]) {
	g._visited[w]=true;
	worklist.push_back(w);
	reaching.push_back(wn2i);
      }
    }
  }

  return false;
}

template<class T, class N2I, class I2NMAP>
void pkh_oto_reorder(std::vector<unsigned int> &reachable,
		   std::vector<unsigned int> &reaching,
		   typename boost::property_map<T, N2I>::type &n2i, 
		   pkh_online_topological_order<T,N2I,I2NMAP> &g) {    
  std::vector<typename T::vertex_descriptor> tmp;
  tmp.reserve(reaching.size() + reachable.size());
  std::vector<unsigned int>::const_iterator iend(reaching.end());
  for(std::vector<unsigned int>::const_iterator i(reaching.begin());
      i!=iend;++i) {
    unsigned int n(g._i2n[*i]);
    tmp.push_back(n);
    g._visited[n]=false;
  }
  iend=reachable.end();
  for(std::vector<unsigned int>::const_iterator i(reachable.begin());
      i!=iend;++i) {
    unsigned int n(g._i2n[*i]);
    tmp.push_back(g._i2n[*i]);
    g._visited[n]=false;
  }
  std::vector<unsigned int>::iterator i(reachable.begin());
  std::vector<unsigned int>::iterator j(reaching.begin());
  std::vector<unsigned int>::iterator jend(reaching.end());
  unsigned int index(0);
  while(i != iend || j != jend) {
    unsigned int w;
    if(j == jend || (i != iend && *i < *j)) {
      w=*i;++i;
    } else {
      w=*j;++j;
    }
    typename T::vertex_descriptor n(tmp[index++]);
    // allocate n at w
    g._i2n[w]=n;
    n2i[n]=w;
  }  
}

template<class T, class N2I, class I2NMAP>
std::pair<typename T::edge_descriptor, bool> 
add_edge(typename pkh_online_topological_order<T,N2I,I2NMAP>::vertex_descriptor t, 
	 typename pkh_online_topological_order<T,N2I,I2NMAP>::vertex_descriptor h, 
	 pkh_online_topological_order<T,N2I,I2NMAP> &g) {

  std::pair<typename T::edge_descriptor, bool> r;
  r = add_edge(static_cast<typename T::vertex_descriptor>(t),
	       static_cast<typename T::vertex_descriptor>(h),
	       static_cast<T&>(g));

  typedef typename boost::property_map<T, N2I>::type N2iMap;
  N2iMap n2i = get(N2I(),g);

  unsigned int hn2i(n2i[h]);
  unsigned int tn2i(n2i[t]);

  if(r.second && hn2i < tn2i) {
    // need to reorder
    std::vector<unsigned int> reaching, reachable;
    if(pkh_oto_dfs<T,N2I,I2NMAP>(h,t,reachable,reaching,n2i,g)) {
      throw std::runtime_error("loop detected");
    } else {
      sort(reaching.begin(),reaching.end());
      sort(reachable.begin(),reachable.end());
      pkh_oto_reorder(reachable,reaching,n2i,g);
    }
  }
}

template<class InputIter, class T, class N2I, class I2NMAP>
void add_edges(InputIter b, InputIter e, 
	       pkh_online_topological_order<T,N2I,I2NMAP> &g) {
  for(;b!=e;++b) {
    add_edge(b->first,b->second,g);
  }
}

template<class T, class N2I = n2i_t, class I2NMAP = std::vector<typename T::vertex_descriptor> >
class pkh_online_topological_order : public T {
public:
  typedef pkh_online_topological_order<T> self;
  friend std::pair<typename T::edge_descriptor, bool> add_edge<T>(typename self::vertex_descriptor, 
								  typename self::vertex_descriptor, 
								  self &);
  friend void pkh_oto_reorder<T,N2I,I2NMAP>(std::vector<unsigned int> &,
					  std::vector<unsigned int> &,
					  typename boost::property_map<T, N2I>::type &n2i, 
					  pkh_online_topological_order<T,N2I,I2NMAP> &g);

  friend bool pkh_oto_dfs<T,N2I,I2NMAP>(typename self::vertex_descriptor h, 
					typename self::vertex_descriptor t,
					std::vector<unsigned int> &reachable,
					std::vector<unsigned int> &reaching,
					typename boost::property_map<T, N2I>::type &n2i,		 
					self &g);
private:
   I2NMAP _i2n;
  std::vector<bool> _visited; // useful and not too expensive.
public:
  pkh_online_topological_order(T const &g, unsigned int acc = 1) 
    : T(g), _visited(num_vertices(g),false) {
    
    // use the topological sort function to
    // order the vertices correctly ...
    _i2n.reserve(num_vertices(g));
    topological_sort(g,std::back_inserter(_i2n));
    // now reverse them
    reverse(_i2n.begin(),_i2n.end());
    // finally, setup the n2i map correctly
    typedef typename boost::property_map<T, N2I>::type N2iMap;
    N2iMap n2i = get(N2I(),*this);
    
    for(unsigned int i=0;i!=_i2n.size();++i) {
      n2i[_i2n[i]]=i;
    }    
  }

  pkh_online_topological_order(typename T::vertices_size_type n,
			       unsigned int acc = 1) 
    : T(n), _i2n(n), _visited(n, false) {
    typedef typename boost::property_map<T, N2I>::type N2iMap;
    N2iMap n2i = get(N2I(),*this);
    
    unsigned int counter(0);
    typename T::vertex_iterator i,iend;
    for(tie(i,iend) = vertices(*this);i!=iend;++i) {
      _i2n[counter]=*i;
      n2i[*i]=counter++;
    }
  }
};

#endif



