// Basically, these algorithms are used to
// generate random graphs ...

#ifndef RANDOM_GRAPH_H
#define RANDOM_GRAPH_H

#include <algorithm>
#if __GNUC__ >= 3
#include <ext/algorithm>
#endif
#include <vector>
#include <stdexcept>
#include <assert.h>
#include <limits.h>

#include <boost/counting_iterator.hpp>
#include <boost/random/geometric_distribution.hpp>
#include <boost/random/random_number_generator.hpp>

// =========================================
// Some Configuration Parameters   
//
#define FAST_RANDOM_DAG_GENERATOR
// #define PRECISE_FAST_RANDOM_DAG_GENERATOR
// =========================================

template<class EdgeList, class UniformRandomNumberGenerator>
void random_edgelist(unsigned int V, double OD, unsigned int L,
		     EdgeList &edges, UniformRandomNumberGenerator &urng) {
  if(L > V) { throw std::runtime_error("invalid locality"); }
  //  double p(((double)(V-OD)) / ((double) V));
  long long L2P1((2*L)+1);
  double p(L2P1);
  p=(p-OD)/p;

  // the idea behind the rorder is to eliminate
  // any pathological advantages
  boost::random_number_generator<UniformRandomNumberGenerator,unsigned int> rng(urng);
  std::vector<unsigned int> rorder;
  std::copy(boost::make_counting_iterator<unsigned int>(0),
	    boost::make_counting_iterator<unsigned int>(V),
	    std::back_inserter(rorder));

  random_shuffle(rorder.begin(),rorder.end(),rng);
 
  if(V == OD) { p = 0.0000000001; } // annoying hack I must say
  boost::geometric_distribution
    <UniformRandomNumberGenerator,unsigned int> gd(urng,p);
  // I subtract one from gd() here because its return value
  // is >= 1, and I want edge zero to be possible.
  assert(((2*L)+1) <= V);
  long long N(gd()-1); 

  while(N < (V*L2P1)) {
    unsigned int t(N / L2P1);
    unsigned int h(N - (t*L2P1));
    // h ranges from 0 to 2L inclusive
    assert(h < L2P1);
    if(h >= L) {
      h = h + t - L;
      
      if(h >= V) {
	// this case is for wrap around
	h -= V;
      }
    } else if(h < L) {
      h = h + t - L;
      
      if(h >= V) {
	// this case is for wrap around
	h += V;
      }
    }

    assert(h < V);

    edges.push_back(typename EdgeList::value_type(rorder[t],rorder[h]));
    N+=gd();
  }
}

double rae_compute_p(unsigned int x, unsigned long long from) {
  double p(from);

  if(x >= from) { 
    return 1.0; 
  } else {
    return (p - x) / p;
  }
}

template<class EdgeList, class UniformRandomNumberGenerator>
  void random_acyclic_edgelist(unsigned int V, unsigned int ssize, 
			       EdgeList &edges, UniformRandomNumberGenerator &urng) {

  // ssize = sample size

  boost::random_number_generator<UniformRandomNumberGenerator,unsigned long long> rng(urng);
  
  // the algorithm doesn't make any guarantee about the
  // distribution of edges in the edgelist. so, using random_shuffle
  // might be a good idea, if that's the desired result
  
  unsigned long long V_VM1_D2(V);
  V_VM1_D2 = (V_VM1_D2*(V_VM1_D2-1)) / 2; 

  // first, generate a random topological order by
  // first initialising the vector such rtorder[i]=i
  // and then shuffling it around.
  std::vector<unsigned int> rtorder;
  std::copy(boost::make_counting_iterator<unsigned int>(0),
	    boost::make_counting_iterator<unsigned int>(V),
	    std::back_inserter(rtorder));

  random_shuffle(rtorder.begin(),rtorder.end(),rng);

#ifdef FAST_RANDOM_DAG_GENERATOR
  boost::geometric_distribution
    <UniformRandomNumberGenerator,unsigned long long> gd(urng,rae_compute_p(ssize,V_VM1_D2));
#endif  

  unsigned long long N(0),Acc(0);
  unsigned long long t(0);

#ifdef FAST_RANDOM_DAG_GENERATOR
  N = gd()-1;
#endif

  while(ssize != 0 && N < V_VM1_D2) {
    while((Acc+V-t-1) <= N) {
      ++t;
      Acc += V - t;
    }
    unsigned long long h(t + (N - Acc) + 1);

#ifdef FAST_RANDOM_DAG_GENERATOR    
    // redetermine probability as we're now choosing ssize records from V_VM1_D2-N edges.
    edges.push_back(typename EdgeList::value_type(rtorder[t],rtorder[h]));
#ifdef PRECISE_FAST_RANDOM_DAG_GENERATOR
    // a slightly more interesting situation
    ssize--;
    gd = boost::geometric_distribution<UniformRandomNumberGenerator,unsigned long long>(urng,rae_compute_p(ssize,V_VM1_D2 - N));
#endif
    N+=gd();
#else 
    if(rng(V_VM1_D2) < ssize) {
      // add edge with probability p
      edges.push_back(typename EdgeList::value_type(rtorder[t],rtorder[h]));
    }
    ++N;
#endif
  }
}


#if 0
// this is basically the above except that I don't adjust
// the priority of the geometric distribution as the edges
// are processed.  This makes it less likely to generate
// a fixed number of edges
template<class EdgeList, class UniformRandomNumberGenerator>
  void random_acyclic_edgelist(unsigned int V, double OD, 
			       EdgeList &edges, UniformRandomNumberGenerator &urng) {

  boost::random_number_generator<UniformRandomNumberGenerator,unsigned int> rng(urng);
  
  // the algorithm doesn't make any guarantee about the
  // distribution of edges in the edgelist. so, using random_shuffle
  // might be a good idea, if that's the desired result
  
  long long V_VM1_D2((V*(V-1)) / 2); 

  // first, generate a random topological order by
  // first initialising the vector such rtorder[i]=i
  // and then shuffling it around.
  std::vector<unsigned int> rtorder;
  std::copy(boost::make_counting_iterator<unsigned int>(0),
	    boost::make_counting_iterator<unsigned int>(V),
	    std::back_inserter(rtorder));

  random_shuffle(rtorder.begin(),rtorder.end(),rng);

  //  double p(((double)(V-OD)) / ((double) V));
  double p(V_VM1_D2);
  OD *= V;
  p=(p-OD)/p;

  boost::geometric_distribution
    <UniformRandomNumberGenerator,unsigned int> gd(urng,p);
  long long N(gd()-1),Acc(0);
  unsigned int t(0);
  
  while(N < V_VM1_D2) {
    while((Acc+V-t-1) <= N) {
      ++t;
      Acc += V - t;
    }
    unsigned int h(t + (N - Acc) + 1);
    edges.push_back(typename EdgeList::value_type(rtorder[t],rtorder[h]));
    N+=gd();
  }
}

#endif

#if 0
template<class EdgeList, class RandomNumberGenerator>
void random_edgelist(unsigned int V, unsigned int E, 
		     EdgeList &edges, RandomNumberGenerator &r) {
  for(unsigned int i=0;i!=V;++i) {
    for(unsigned int j=0;j!=V;++j) {
      if(r(V) < E) {
	edges.push_back(typename EdgeList::value_type(i,j));
      }
    }
  }
}

#endif

#if 0

template<class EdgeList, class RandomNumberGenerator>
void random_edgelist(unsigned int V, unsigned int E, 
		     EdgeList &edges, RandomNumberGenerator &r) {
  // the algorithm doesn't make any guarantee about the
  // distribution of edges in the edgelist. so, using random_shuffle
  // might be a good idea, if that's the desired result

  unsigned int Vsqr(V*V);  

  assert(V != 0);
  assert(E < Vsqr);

  // obtain E random samples from the enumeration of edges
  // note: unfortunately this is quadratic in V. this is a shame
  // as it would be nice for random_sample_n to do better...
  std::vector<unsigned int> sample;
  random_sample_n(boost::make_counting_iterator<unsigned int>(0),
		       boost::make_counting_iterator<unsigned int>(Vsqr),
		       std::back_inserter(sample),E,r);

  // ASSERT: sample[i] < sample[j] iff i < j, by defn of random_sample_n

  std::vector<unsigned int>::iterator iend(sample.end());
  for(std::vector<unsigned int>::iterator i(sample.begin());
      i!=iend;++i) {
    unsigned int t(*i / V);
    unsigned int h(*i - (t*V));
    edges.push_back(typename EdgeList::value_type(t,h));
  }


  multiset<unsigned int> randseq;
  
  for(unsigned int i=0;i!=E;++i) {
    unsigned int tmp(r(Vsqr - i));
    randseq.insert(i);    
  }
}

template<class EdgeList, class RandomNumberGenerator>
  void random_fast_acyclic_edgelist(unsigned int V, unsigned int E, 
				    EdgeList &edges, RandomNumberGenerator &r) {
  
  // the algorithm doesn't make any guarantee about the
  // distribution of edges in the edgelist. so, using random_shuffle
  // might be a good idea, if that's the desired result
  
  unsigned int V_VM1_D2((V*(V-1)) / 2); 

  assert(V != 0);
  assert(E < V_VM1_D2);
  
  // first, generate a random topological order by
  // first initialising the vector such rtorder[i]=i
  // and then shuffling it around.
  std::vector<unsigned int> rtorder;
  std::copy(boost::make_counting_iterator<unsigned int>(0),
	    boost::make_counting_iterator<unsigned int>(V),
	    std::back_inserter(rtorder));

  random_shuffle(rtorder.begin(),rtorder.end(),r);
  
  // obtain E random samples from the enumeration of edges
  std::vector<unsigned int> sample;
  random_sample_n(boost::make_counting_iterator<unsigned int>(0),
		  boost::make_counting_iterator<unsigned int>(V_VM1_D2),
		  std::back_inserter(sample),E,r);

  // ASSERT: sample[i] < sample[j] iff i < j, by defn of random_sample_n

  unsigned int N(0),t(0);
  std::vector<unsigned int>::iterator iend(sample.end());
  for(std::vector<unsigned int>::iterator i(sample.begin());
      i!=iend;++i) {
    assert(*i < V_VM1_D2);
    while(*i >= (N+V-t-1)) {
      ++t;
      N += V-t;
    }
    unsigned int h(t + (*i - N) + 1);
    assert(t != h);
    edges.push_back(typename EdgeList::value_type(rtorder[t],rtorder[h]));
  }
}

#endif

#endif
