// Copyright @ David J. Pearce, May 2005.
//
// Author: David James Pearce
// Email:  d.pearce@doc.ic.ac.uk
// Date:   May 2003
//
// Notes
// =====
// This is an extension (by me) of the following work to
// compute the strong components of a digraph online:
//
// "On-Line Graph Algorithms for Incremental Compilation", 
// Alberto Marchetti-Spaccamela, Umberto Nanni and Hans 
// Rohnert. Information Processing Letters, 1996.

#ifndef SNRP_ONLINE_STRONG_COMPONENTS_HPP
#define SNRP_ONLINE_STRONG_COMPONENTS_HPP

#include <stdexcept>
#include <boost/property_map.hpp>
#include <boost/graph/topological_sort.hpp>

using namespace std;
using namespace boost;

#include "osc_tags.hpp"

template<class T, class C2I, class I2CMAP>
class snrp_online_strong_components;

template<class T, class C2I, class I2CMAP>
class snrp_osc_algo {
private:
  typedef typename T::vertex_descriptor vd_t;
  typedef typename T::out_edge_iterator out_iterator;
public:
  unsigned int lb;
  unsigned int ub;
  vector<bool> visited;
  vector<bool> in_component;
  property_map<T,C2I>::type &c2i;
  snrp_online_strong_components<T,C2I,I2CMAP> &g
public:
  snrp_osc_dfs(unsigned int _lb, unsigned int _ub,
	       snrp_online_strong_components<T,C2I,I2CMAP> &_g) 
    : lb(_lb), ub(_lb), visited((_ub-_lb)+1,false),
      in_component((_ub-_lb)+1,false), c2i(get(C2I(),_g)), g(_g) {

    visited[ub-lb]=true;
    in_component[ub-lb]=true;
  }

#define VISITED(x) visited[c2i[x]-lb]
#define IN_COMPONENT(x) in_component[c2i[x]-lb]
  
  void dfs(vd_t n) {
    VISITED(n)=true;    
    
    out_iterator i,iend;
    for(tie(i,iend) = out_edges(n);i!=iend;++i) {
      assert(c2i[*i] >= _lb);
      if(c2i[*i] <= _ub) {
	if(!VISITED(*i)) {
	  dfs(*i);
	}
	IN_COMPONENT(n) = IN_COMPONENT(n) | IN_COMPONENT(*i);
      }
    }     
  }

  void shift(void) {
    
  }

#undef VISITED
#undef IN_COMPONENT
};

template<class T, class C2I, class I2CMAP>
std::pair<typename T::edge_descriptor, bool> 
add_edge(typename snrp_online_strong_components<T,C2I,I2CMAP>::vertex_descriptor t, 
	 typename snrp_online_strong_components<T,C2I,I2CMAP>::vertex_descriptor h, 
	 snrp_online_strong_components<T,C2I,I2CMAP> &g) {

  std::pair<typename T::edge_descriptor, bool> r;
  r = add_edge(static_cast<typename T::vertex_descriptor>(t),
	       static_cast<typename T::vertex_descriptor>(h),
	       static_cast<T&>(g));

  typedef property_map<T, C2I>::type C2iMap;
  C2iMap C2i = get(C2I(),g);

  unsigned int lb(c2i[h]);
  unsigned int ub(c2i[t]);

  if(r.second && lb < ub) {
    // need to reorder
    snrp_osc_algo algo(lb,ub,*this);
    algo.dfs(h);
    algo.shift();
  }

  return r;
}

template<class T, class C2I = n2i_t, class I2CMAP = vector<typename T::vertex_descriptor> >
class snrp_online_strong_components : public T {
public:
  typedef snrp_online_strong_components<T> self;
  friend std::pair<typename T::edge_descriptor, bool> add_edge<T>(typename self::vertex_descriptor, 
								  typename self::vertex_descriptor, 
								  self &);
  friend class snrp_oto_algo<T,C2I,I2CMAP>;
private:
  I2CMAP _i2c;
public:
  snrp_online_strong_components(T const &g) 
    : T(g) {
    
    // use the topological sort function to
    // order the vertices correctly ...
    topological_sort(g,std::back_inserter(_i2c));
    // now reverse them
    reverse(_i2c.begin(),_i2c.end());
    // finally, setup the c2i map correctly
    typedef property_map<T, C2I>::type C2iMap;
    C2iMap c2i = get(C2I(),*this);
    
    for(unsigned int i=0;i!=_i2c.size();++i) {
      c2i[_i2c[i]]=i;
    }    
  }

  snrp_online_strong_components(typename T::vertices_size_type n) 
    : T(n), _i2c(n) {
    typedef property_map<T, C2I>::type C2iMap;
    C2iMap C2i = get(C2I(),*this);
    
    unsigned int counter(0);
    typename T::vertex_iterator i,iend;
    for(tie(i,iend) = vertices(*this);i!=iend;++i) {
      _i2c[counter]=*i;
      c2i[*i]=counter++;
    }
  }
};

#endif



