// A test for the topological graph orders
// =======================================
// Author: David J. Pearce, 2003
//
// the idea behind this test is twofold:
//
// 1) we want to make sure that the topological
//    orders give correct answers
// 2) we want to compare the speeds of the different
//    algorithms.

#include <iostream>
#include <cstdlib>
#include <getopt.h>
#include <sys/time.h>

#include <set>
#include <vector>
#include <algorithm>
#include <numeric>

// unsure why this is needed
#define BOOST_NO_STDC_NAMESPACE

#include <boost/random.hpp>
#include <boost/graph/adjacency_list.hpp>
#include <boost/graph/depth_first_search.hpp>
#include <boost/graph/copy.hpp>
#include <boost/lexical_cast.hpp>
#include "booste/random_graph.hpp"
#include "booste/path.hpp"
#include "booste/mnr_online_topological_order.hpp"
#include "booste/pkh_online_topological_order.hpp"
#include "booste/ahrsz_online_topological_order.hpp"
#include "booste/ahrsz_online_topological_order2.hpp"
#include "booste/batch_mnr_online_topological_order.hpp"
#include "booste/simple_topological_order.hpp"

class my_timer {
private:
  struct timeval _start;
public:
  my_timer(void) {
    gettimeofday(&_start,NULL);
  }

  double elapsed(void) {
    struct timeval tmp;
    gettimeofday(&tmp,NULL); 
    double end = tmp.tv_sec + (tmp.tv_usec / 1000000.0);
    double start = _start.tv_sec + (_start.tv_usec / 1000000.0);    
    return end - start;
  }  
};

using namespace std;
using namespace boost;

typedef adjacency_list<setS,vecS, bidirectionalS, property<n2i_t,unsigned int> > subgraph1_t;
typedef pkh_online_topological_order<subgraph1_t> graph1_t;
typedef adjacency_list<setS,vecS, bidirectionalS, property<n2i_t,unsigned int> > subgraph2_t;
typedef mnr_online_topological_order<subgraph2_t> graph2_t;
typedef adjacency_list<setS,vecS, bidirectionalS, property<n2i_t,ahrsz_priority_value<> > > subgraph3_t;
typedef ahrsz_online_topological_order<subgraph3_t> graph3_t;
typedef adjacency_list<setS,vecS, bidirectionalS, property<n2i_t,ahrsz_priority_value<ordered_slist<void> > > > subgraph4_t;
typedef ahrsz_online_topological_order<subgraph4_t,ordered_slist<void> > graph4_t;
typedef adjacency_list<setS,vecS, bidirectionalS, property<n2i_t,experimental::ahrsz_priority_value> > subgraph5_t;
typedef experimental::ahrsz_online_topological_order<subgraph5_t> graph5_t;
typedef adjacency_list<setS,vecS, bidirectionalS, property<n2i_t,unsigned int> > subgraph6_t;
typedef batch_mnr_online_topological_order<subgraph6_t> graph6_t;
typedef adjacency_list<setS,vecS, bidirectionalS, property<n2i_t,unsigned int> > subgraph7_t;
typedef simple_topological_order<subgraph7_t> graph7_t;


#define MAJOR_VERSION 1
#define MINOR_VERSION 3

#define OPT_HELP 0
#define OPT_VERSION 1
#define OPT_NODES 2
#define OPT_CHECKING 3
#define OPT_RATIO 6
#define OPT_MNR 7
#define OPT_OFFLINE 8
#define OPT_QUIET 9
#define OPT_CYCLES 10
#define OPT_OUTDEGREE 12
#define OPT_AHRSZ 13
#define OPT_AHRSZE 14
#define OPT_ACC 15
#define OPT_PKH 16
#define OPT_COST 17
#define OPT_RTXY 18
#define OPT_ARXY 19
#define OPT_OVER 20
#define OPT_DXY 21
#define OPT_AHRSZB 22
#define OPT_DDXY 23
#define OPT_BMNR 24
#define OPT_BATCH 25

#define PKH 0
#define MNR 1
#define AHRSZ 2
#define AHRSZB 3
#define AHRSZE 4
#define OFFLINE 5
#define BATCH_MNR 6

template<class T>
void my_print_graph(T &graph) {
  typedef typename T::out_edge_iterator oiterator;

  cout << "\tgraph = { ";
  typename T::vertex_iterator i,iend;
  for(tie(i,iend) = vertices(graph);i!=iend;++i) {
    oiterator j,jend;
    for(tie(j,jend) = out_edges(*i,graph);
	j!=jend;++j) {
      cout << source(*j,graph) << "->" << target(*j,graph) << " ";
    }    
  }
  cout << " }" << endl;
}

template<class S>
class n2i_comp {
private:
  typedef typename S::vertex_descriptor vd_t;
  typedef typename property_map<S, n2i_t>::type N2iMap;
  N2iMap _n2i;
public:
  n2i_comp(N2iMap n) : _n2i(n) {
  }

  bool operator()(vd_t const &v1, vd_t const &v2) {
    return _n2i[v1] < _n2i[v2];
  }
};

template<class T, class S>
void print_order(T &graph) {
  typedef typename T::vertex_descriptor vd_t;
  typedef typename property_map<S, n2i_t>::type N2iMap;
  multiset<vd_t, n2i_comp<S> > ord(get(n2i_t(),graph));

  cout << "\torder = { ";
  typename T::vertex_iterator i,iend;
  for(tie(i,iend) = vertices(graph);i!=iend;++i) {
    ord.insert(*i);
  }
  for(typename multiset<vd_t, n2i_comp<S> >::iterator i(ord.begin());
      i!=ord.end();++i) {
    cout << *i << " ";
  }
  cout << " }" << endl;
}

template<class T, class S>
unsigned int compute_size_dxy(unsigned int t, unsigned int h, double &ddxy, T &g) {
  typedef typename property_map<S, n2i_t>::type N2iMap;
  typedef typename T::vertex_descriptor vd_t;
  typedef typename T::out_edge_iterator out_iterator;
  typedef typename T::in_edge_iterator in_iterator;

  N2iMap n2i = get(n2i_t(),g);

  unsigned int dxy(0);
  vector<vd_t> worklist;    
  vector<bool> visited(num_vertices(g),false);

  // mark node h as visited
  
  visited[h]=true;
  worklist.push_back(h);
  ++ddxy; // for h

  // now perform the forward depth-first-search
  while(worklist.size() > 0) { 
    vd_t n(worklist.back());
    worklist.pop_back();
    dxy++;

    out_iterator i,iend;
    for(tie(i,iend) = out_edges(n,g); i!=iend;++i) {
      unsigned int w(target(*i,g));
      ++ddxy; // for w
      if(!(n2i[w] > n2i[t]) && !visited[w]) {
	visited[w]=true;
	worklist.push_back(w);
      }
    }
  }

  // Now, perform the reverse depth-first-search
  worklist.push_back(t);
  visited[t] = true;
  ++ddxy; // for t

  while(worklist.size() > 0) { 
    vd_t n(worklist.back());
    worklist.pop_back();
    dxy++;

    in_iterator i,iend;
    for(tie(i,iend) = in_edges(n,g); i!=iend;++i) {
      unsigned int w(source(*i,g));
      ++ddxy; // for w
       if(!(n2i[w] < n2i[h]) && !visited[w]) {
	visited[w]=true;
	worklist.push_back(w);
      }
    }
  }

  return dxy;
}

template<class T, class S>
unsigned int compute_size_cones(unsigned int t, unsigned int h, T &g) {
  typedef typename property_map<S, n2i_t>::type N2iMap;
  typedef typename T::vertex_descriptor vd_t;
  typedef typename T::out_edge_iterator out_iterator;
  typedef typename T::in_edge_iterator in_iterator;

  N2iMap n2i = get(n2i_t(),g);

  unsigned int dxy(0);
  vector<vd_t> worklist;    
  vector<bool> visited(num_vertices(g),false);
    
  // mark node h as visited
  
  visited[h]=true;
  worklist.push_back(h);

  // now perform the forward depth-first-search
  while(worklist.size() > 0) { 
    vd_t n(worklist.back());
    worklist.pop_back();
    dxy++;

    out_iterator i,iend;
    for(tie(i,iend) = out_edges(n,g); i!=iend;++i) {
      unsigned int w(target(*i,g));
      if(!visited[w]) {
	visited[w]=true;
	worklist.push_back(w);
      }
    }
  }

  // Now, perform the reverse depth-first-search
  worklist.push_back(t);
  visited[t] = true;

  while(worklist.size() > 0) { 
    vd_t n(worklist.back());
    worklist.pop_back();
    dxy++;

    in_iterator i,iend;
    for(tie(i,iend) = in_edges(n,g); i!=iend;++i) {
      unsigned int w(source(*i,g));
       if(!visited[w]) {
	visited[w]=true;
	worklist.push_back(w);
      }
    }
  }

  return dxy;
}

template<class T, class S>
bool check_solution(T &graph, string const &s) {
  typedef typename property_map<S, n2i_t>::type N2iMap;
  N2iMap n2i = get(n2i_t(),graph);

  typename T::vertex_iterator i,iend;
  for(tie(i,iend) = vertices(graph);i!=iend;++i) {
    typename T::vertex_iterator j(i),jend,jdum;
    ++j;
    for(tie(jdum,jend) = vertices(graph);j!=jend;++j) {
      bool f1(n2i[*i] < n2i[*j]);
      bool f2(n2i[*j] < n2i[*i]);
      if(f1 && path(*j,*i,graph)) {
	cerr << "Check failure because n2i[" << *i << "] < n2i[" << *j << "] AND path("; 
	cerr << *j << "," << *i << "). " << s << endl;
	my_print_graph(graph);
	print_order<T,S>(graph);
	return false;
      } else if(f2 && path(*i,*j,graph)) {
	cerr << "Check failure because n2i[" << *j << "] < n2i[" << *i << "] AND path("; 
	cerr << *i << "," << *j << "). " << s << endl;
	my_print_graph(graph);
	print_order<T,S>(graph);
      } else if(!f1 && !f2 && (path(*i,*j,graph) || path(*j,*i,graph))) {
	cerr << "Check failure because n2i[" << *j << "] == n2i[" << *i;
	cerr << "] AND there is at least one path connected them. " << s << endl;
	my_print_graph(graph);
	print_order<T,S>(graph);
      }
    }
  }
  return true;
}

class work_visitor : public default_dfs_visitor {
protected:
  vector<set<unsigned int> > &_scores;
public:
  work_visitor(vector<set<unsigned int> > &scores) 
    : _scores(scores) {
  }

  template < typename Vertex, typename Graph >
  void finish_vertex(Vertex u, const Graph & g) const {
    // at this point we must compute our own score
    _scores[u].insert(u);
    typename Graph::out_edge_iterator i,iend;
    for(tie(i,iend)=out_edges(u,g);i!=iend;++i) {
      _scores[u].insert(_scores[target(*i,g)].begin(),_scores[target(*i,g)].end());
    }
  }
};

class workload {
public:
  class workload_mode {
  public:
    bool checking;
    bool cost;
    bool RTXY;
    bool ARXY;
    bool DXY;
    bool DDXY;
    workload_mode(void)
      : checking(false), cost(false),
	RTXY(false), ARXY(false),DXY(false) {
    }
  } mode;
  
  bool acyclic;
  unsigned int N;
  unsigned int E;
  unsigned int over;
  unsigned int batch;
  
  workload(void) 
    : acyclic(true), N(50), E(0), 
    over(1000), batch(1) {
  }
};

template<class T, class S, class UniformRandomNumberGenerator>
void build_graph(unsigned int N, unsigned int initial, unsigned int over, 
		 vector<pair<unsigned int, unsigned int> > &spare, 
		 deque<pair<unsigned int, unsigned int> > &used,
		 T &graph,
		 UniformRandomNumberGenerator &urng) {

  typedef adjacency_list<setS,vecS, directedS> tmpgraph_t;
  // get normal random_number_generator.
  boost::random_number_generator<UniformRandomNumberGenerator,unsigned int> rng(urng);
  // generate random edgelist
  random_acyclic_edgelist(N,initial+over,spare,urng);
  // generate random permutation of edge list
  random_shuffle(spare.begin(),spare.end(),rng);
  // at this point we add enough edges to get us upto IOD
  S tmpg(N);
  
  while(initial-- > 0) {
    pair<unsigned int, unsigned int> &e(spare.back());    
    add_edge(e.first, e.second, tmpg);
    used.push_back(e);
    spare.pop_back();
  }

  assert(num_vertices(tmpg) == N);
  graph = T(tmpg);
}
		 
template<class T, class S, class URNG>
void do_work(workload &w, URNG &urng) {
  vector<pair<unsigned int, unsigned int> > spare; // unused edges
  deque<pair<unsigned int, unsigned int> > used;  // edges currently part of graph
  unsigned int errors(0);
  T graph(w.N); 
  // generate a random DAG.

  build_graph<T,S,URNG>(w.N, w.E, w.over, 
	      spare, used, graph, urng);

  // get a random number generator
  boost::random_number_generator<URNG,unsigned int> rng(urng);
  unsigned int num_processed(spare.size()), num_invalidating(0);
  double tRTXY(0),tARXY(0),tDXY(0),tDDXY(0);

  typedef typename property_map<S, n2i_t>::type N2iMap;
  N2iMap n2i = get(n2i_t(),graph);

  if(w.mode.checking) {
    if(!check_solution<T,S>(graph,"Initial Graph")) {
      errors ++;
    }
  }

  // begin timing
  my_timer t; 

  while(spare.size() > 0) {
    // add a random edge
    vector<pair<unsigned int, unsigned int> >::iterator i;
    if(spare.size() < w.batch) {
      i = spare.begin();
    } else {
      i = spare.end() - w.batch;
    }

#if 0
    if(w.mode.ARXY || w.mode.DXY || w.mode.RTXY) {
      if(n2i[e.second] < n2i[e.first]) {
	if(w.mode.ARXY) {
	  typename T::vertex_iterator i,iend;
	  for(tie(i,iend) = vertices(graph);i!=iend;++i) {
	    if(!(n2i[e.second] > n2i[*i]) && !(n2i[*i] > n2i[e.first])) {
	      tARXY++;
	    }
	  }
	}

	if(w.mode.DXY || w.mode.DDXY) {
	  tDXY += compute_size_dxy<T,S>(e.first,e.second,tDDXY,graph);
	}

	if(w.mode.RTXY) {
	  tRTXY += compute_size_cones<T,S>(e.first,e.second,graph);
	}

	num_invalidating++;
      }
    }
#endif    

    // simple to start with ...
    add_edges(i,spare.end(),graph);
    used.insert(used.end(),i,spare.end());
    // now, select count random edges which are in graph and delete them
    // so as to maintain perfect outdegree ...
    unsigned int count(spare.end()-i);
    while(count-- > 0) {
      assert(used.size() > 0);
      unsigned int index = rng(used.size());
      remove_edge(used[index].first,used[index].second,graph);
      swap(*(used.begin()+index),*(used.end()-1));
      used.pop_back();
    }
    
    if(w.mode.checking) {
      if(!check_solution<T,S>(graph,"BATCH")) {
	errors++;
      }
    }
 
    spare.erase(i,spare.end());
  }

  if(errors > 0) {
    cerr << errors << " errors." << endl;
  }

  if(w.mode.cost) {
    cout << t.elapsed() / num_processed << "\t";
  }
  if(w.mode.ARXY) {
    cout << tARXY / num_invalidating << "\t";
  }
  if(w.mode.RTXY) {
    cout << tRTXY / num_invalidating << "\t";
  }
  if(w.mode.DXY) {
    cout << tDXY / num_invalidating << "\t";
  }
  if(w.mode.DDXY) {
    cout << tDDXY / num_invalidating << "\t";
  }
  cout << num_processed << "\t" << num_invalidating;
}

int main(int argc, char *argv[]) {

  struct option long_options[]={
    {"help",no_argument,NULL,OPT_HELP},
    {"version",no_argument,NULL,OPT_VERSION},
    {"checking",no_argument,NULL,OPT_CHECKING},
    {"nodes",required_argument,NULL,OPT_NODES},
    {"outdegree",required_argument,NULL,OPT_OUTDEGREE},
    {"MNR",no_argument,NULL,OPT_MNR},
    {"BATCH-MNR",no_argument,NULL,OPT_BMNR},
    {"AHRSZ",no_argument,NULL,OPT_AHRSZ},
    {"PKH",no_argument,NULL,OPT_PKH},
    {"AHRSZb",no_argument,NULL,OPT_AHRSZB},
    {"ahrsze",no_argument,NULL,OPT_AHRSZE},
    {"offline",no_argument,NULL,OPT_OFFLINE},
    {"allow-cycles",no_argument,NULL,OPT_CYCLES},
    {"accumulate",required_argument,NULL,OPT_ACC},
    {"over",required_argument,NULL,OPT_OVER},
    {"cost",no_argument,NULL,OPT_COST},
    {"RTXY",no_argument,NULL,OPT_RTXY},
    {"ARXY",no_argument,NULL,OPT_ARXY},
    {"DXY",no_argument,NULL,OPT_DXY},
    {"DDXY",no_argument,NULL,OPT_DDXY},
    {"batch",required_argument,NULL,OPT_BATCH},
    NULL
  };

  char *descriptions[]={
    " -h     --help                    display this information",
    " -v     --version                 display version information",
    "        --checking                perform correctness checking",
    " -n<x>  --nodes=<x>               set value of N (default=50)",
    " -o<x>  --over=<x>                compute values (e.g. cost) over x insertions",
    " -b<x>  --batch=<x>               process edges in x chunks. This is useful for",
    "                                  looking at the non-unit change algorithms",
    " -c     --cost                    find average cost.",
    "        --RTXY                    compute average size of reach trees.",
    "                                  Note! Inteferes with cost measurement.",
    "        --ARXY                    compute average Affected Region size.",
    "                                  Note! Inteferes with cost measurement.",
    "        --DXY                     compute average delta_xy size.",
    "                                  Note! Inteferes with cost measurement.",
    "        --DDXY                    compute average delta_xy extended size.",
    "                                  Note! Inteferes with cost measurement.",
    " -d<x>  --outdegree=<x>           set expected initial outdegree",
    "                                  determintes the probility p to use in the graph",
    "                                  generator and does not *guarantee* an exact outdegree",
    "                                  of x",
    "                                  as a percentage. (default=50)",
    "        --allow-cycles            permit cycles in the graph",
    "        --MNR                     use online topological sorting algorithm by",
    "                                  Marchetti-Spaccamela, Nanni and Rohnert",
    "        --PKH                     use online topological order algorithm by",
    "                                 Pearce, Kelly, Hankin",
    "        --AHRSZ                   Use algorithm by Alpern et al.  This implementation uses",
    "                                  an O(1) amortized time priority space data structure",
    "        --AHRSZb                  Use algorithm by Alpern et al.  This implementation uses",
    "                                  an O(log n) time priority space data structure",
    "        --AHRSZE                  experimental optimised version",
    "        --BATCH-MNR               use batch online topological ordering algorithm by",
    "                                  David J. Pearce, which is based on the MNR algorithm",

    "",
    "notes: the output is the time measured in seconds, using gettimeofday.",
    NULL
  };

  bool upqf(false);
  unsigned int algorithm(PKH);
  unsigned int over(0);
  double outdegree=0;
  workload work;

  typedef boost::mt19937 random_gen_t;
  //typedef boost::rand48 random_gen_t;

  random_gen_t rgen;
  //  boost::random_number_generator<random_gen_t,unsigned int> rgen(_rgen);

  // seed the random number generator
  struct timeval tmp;
  gettimeofday(&tmp,NULL);
  rgen.seed((tmp.tv_sec * 1000000) + tmp.tv_usec); 

  // parse command line arguments
  char v;
  while((v=getopt_long(argc,argv,"b:hvcn:d:o:",long_options,NULL)) != -1) {
    switch(v)
      {      
      case 'h':
      case OPT_HELP:
	cout << "usage: " << argv[0] << " [options]" << endl;
	cout << "options:" << endl;
	for(char **ptr=descriptions;*ptr != NULL; ptr++) {
	  cout << *ptr << endl;
	}    
	exit(1);
      case 'v':
      case OPT_VERSION:
	cout << argv[0] << " v" << MAJOR_VERSION << "." << MINOR_VERSION << endl;
	cout << "\tWritten by David J. Pearce, Feburary 2003" << endl;
	exit(1);
      case OPT_CHECKING:
	work.mode.checking = true;
	break;
      case 'n':
      case OPT_NODES:
	work.N=atoi(optarg);
	break;
      case 'c':
      case OPT_COST:
	work.mode.cost = true;
	break;
      case 'd':
      case OPT_OUTDEGREE:
	outdegree=strtod(optarg,NULL);
	break;
      case OPT_MNR:
	algorithm=MNR;
	break;
      case OPT_BMNR:
	algorithm=BATCH_MNR;
	break;
      case OPT_AHRSZ:
	algorithm=AHRSZ;
	break;
      case OPT_AHRSZB:
	algorithm=AHRSZB;
	break;
      case OPT_AHRSZE:
	algorithm=AHRSZE;
	break;
      case OPT_CYCLES:
	work.acyclic=false;
	break;
      case OPT_RTXY:
	work.mode.RTXY = true;
	break;
      case OPT_ARXY:
	work.mode.ARXY = true;
	break;
      case OPT_DXY:
	work.mode.DXY = true;
	break;
      case OPT_DDXY:
	work.mode.DDXY = true;
	break;
      case 'o':
      case OPT_OVER:
	over = atoi(optarg);
	break;
      case OPT_PKH:
	algorithm=PKH;
	break;
      case 'b':
      case OPT_BATCH:
	work.batch = atoi(optarg);
	break;
      case OPT_OFFLINE:
	algorithm=OFFLINE;
	break;
      default:
	// getopt will print invalid argument for us
	exit(1);
      }
  }

  outdegree *= work.N;
  work.E = (unsigned int) outdegree;

  if(over != 0) {
    work.over = over;
  }

  try {
    if(algorithm == PKH) {
      do_work<graph1_t,subgraph1_t>(work,rgen);
    } else if(algorithm == MNR) {
      do_work<graph2_t,subgraph2_t>(work,rgen);
    } else if(algorithm == AHRSZ) {
      do_work<graph3_t,subgraph3_t>(work,rgen);
    } else if(algorithm == AHRSZB) {
      do_work<graph4_t,subgraph4_t>(work,rgen);
    } else if(algorithm == AHRSZE) {
      do_work<graph5_t,subgraph5_t>(work,rgen);
    } else if(algorithm == BATCH_MNR) {
      do_work<graph6_t,subgraph6_t>(work,rgen);
    } else {
      do_work<graph7_t,subgraph7_t>(work,rgen);
    }
  } catch(runtime_error &e) {
    cout << "Internal failure - " << e.what() << endl;
    exit(1);
  }
  exit(0);
}








